from typing import Any
import aiohttp
import feedparser
from housaku.utils import clean_html


async def fetch_feed(client: aiohttp.ClientSession, feed_url: str) -> list[Any]:
    resp = await client.get(feed_url)
    if resp.status >= 400:
        raise Exception(f"something went wrong while fetching '{feed_url}'")

    content = await resp.text()
    d = feedparser.parse(content)
    if d.bozo:
        raise Exception(f"failed to parse '{feed_url}'")

    return [entry for entry in d.entries]


async def fetch_post(client: aiohttp.ClientSession, post_url: str) -> str:
    resp = await client.get(post_url)
    if resp.status >= 400:
        raise Exception(f"something went wrong while fetching '{post_url}'")

    html = await resp.text()
    cleaned_html = clean_html(html)
    return cleaned_html



from math import log
from sqlalchemy import func, select
from sqlalchemy.orm.session import Session
from housaku.models import Doc, Posting, Word


def search(session: Session, tokens: list[str], limit: int | None = None) -> list[Doc]:
    num_docs = session.scalar(select(func.count()).select_from(Doc))
    if not num_docs:
        return []

    postings = session.scalars(
        select(Posting).join(Word).where(Word.word.in_(tokens))
    ).all()

    df_counts = session.execute(
        select(Word.word, func.count(Posting.doc_id))
        .join(Posting)
        .where(Word.word.in_(tokens))
        .group_by(Word.word)
    ).fetchall()

    idf_values = {word: log(num_docs / (df if df > 0 else 1)) for word, df in df_counts}

    tf_idf_scores = {}
    for posting in postings:
        word = posting.word.word
        tf_idf = posting.tf * idf_values.get(word, 0)
        tf_idf_scores[posting.doc_id] = tf_idf_scores.get(posting.doc_id, 0) + tf_idf

    sorted_doc_ids = sorted(
        tf_idf_scores.keys(),
        key=lambda doc_id: tf_idf_scores[doc_id],
        reverse=True,
    )

    stmt = select(Doc).where(Doc.id.in_(sorted_doc_ids))
    if limit:
        stmt = stmt.limit(limit)

    documents = session.scalars(stmt).all()

    return list(documents)



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from housaku.repositories import DocRepository
from housaku.models import Base, Doc


@pytest.fixture(scope="function")
def session():
    engine = create_engine("sqlite:///:memory:")
    Base.metadata.create_all(engine)

    with Session(engine) as session:
        yield session


@pytest.fixture
def docs_repository(session):
    return DocRepository(session)


@pytest.fixture
def new_doc():
    return Doc(uri="./guide.txt", content="Don't Panic.", properties={})


def test_add_document(docs_repository, new_doc):
    doc_in_db = docs_repository.add(new_doc)

    assert doc_in_db is not None
    assert doc_in_db.id == 1
    assert doc_in_db.content == new_doc.content


def test_get_document_by_id(docs_repository, new_doc):
    docs_repository.add(new_doc)

    doc_in_db = docs_repository.get(id=1)
    assert doc_in_db is not None
    assert doc_in_db.id == 1
    assert doc_in_db.content == new_doc.content


def test_get_document_by_attributes(docs_repository, new_doc):
    docs_repository.add(new_doc)

    doc_in_db = docs_repository.get_by_attributes(uri=new_doc.uri)
    assert doc_in_db is not None
    assert doc_in_db.id == 1
    assert doc_in_db.content == new_doc.content


def test_get_all_documents(docs_repository):
    for i in range(5):
        docs_repository.add(Doc(uri=f"{i}.txt", content="w" * i, properties={}))

    docs_in_db = docs_repository.get_all()
    assert len(docs_in_db) == 5


def test_update_document(docs_repository, new_doc):
    doc_in_db = docs_repository.add(new_doc)
    assert doc_in_db is not None
    assert doc_in_db.content == new_doc.content
    old_content = doc_in_db.content

    docs_repository.update(doc_in_db.id, content="Maybe panic a little.")
    updated_doc = docs_repository.get(doc_in_db.id)
    assert updated_doc is not None
    assert updated_doc.content != old_content


def test_delete_document(docs_repository, new_doc):
    doc_in_db = docs_repository.add(new_doc)
    assert doc_in_db is not None

    deleted_doc = docs_repository.delete(doc_in_db.id)
    assert deleted_doc is not None

    doc_in_db = docs_repository.get(id=doc_in_db.id)
    assert doc_in_db is None



import pytest
from pathlib import Path
from housaku.files import list_files, read_docx, read_epub, read_txt, read_md, read_pdf

TEST_FILES_DIR = Path(__file__).parent / "examples"


def test_list_files():
    file_list = set()
    for file in list_files(TEST_FILES_DIR):
        file_list.add(file)

    assert len(file_list) == sum(
        1 for item in TEST_FILES_DIR.iterdir() if item.is_file()
    )


@pytest.mark.asyncio
async def test_read_txt():
    for test_file in TEST_FILES_DIR.glob("*.txt"):
        tokens, metadata = await read_txt(test_file)
        assert len(tokens)
        assert metadata


@pytest.mark.asyncio
async def test_read_md():
    for test_file in TEST_FILES_DIR.glob("*.md"):
        tokens, metadata = await read_md(test_file)
        assert len(tokens)
        assert metadata


@pytest.mark.asyncio
async def test_read_pdf():
    for test_file in TEST_FILES_DIR.glob("*.pdf"):
        tokens, metadata = await read_pdf(test_file)
        assert len(tokens)
        assert metadata


@pytest.mark.asyncio
async def test_read_epub():
    for test_file in TEST_FILES_DIR.glob("*.epub"):
        tokens, metadata = await read_epub(test_file)
        assert len(tokens)
        assert metadata


@pytest.mark.asyncio
async def test_read_docx():
    for test_file in TEST_FILES_DIR.glob("*.docx"):
        tokens, metadata = await read_docx(test_file)
        assert len(tokens)
        assert metadata


@pytest.mark.asyncio
async def test_bench_read_txt(benchmark):
    test_file = TEST_FILES_DIR / "gutenberg_moby_dick.txt"
    tokens, _ = await benchmark(read_txt, test_file)
    assert tokens


@pytest.mark.asyncio
async def test_bench_read_md(benchmark):
    test_file = TEST_FILES_DIR / "daring_fireball_markdown_syntax.md"
    tokens, _ = await benchmark(read_md, test_file)
    assert tokens


@pytest.mark.asyncio
async def test_bench_read_pdf(benchmark):
    test_file = TEST_FILES_DIR / "gutenberg_the_modern_prometheus.pdf"
    tokens, _ = await benchmark(read_pdf, test_file)
    assert tokens



import pytest
from housaku.utils import tokenize


@pytest.fixture
def short_text() -> str:
    return """
    Far out in the uncharted backwaters of the unfasionable end of
    the western spiral arm of the Galaxy lies a small, unregarded yellow sun.
    """


@pytest.fixture
def medium_text() -> str:
    return """
    A towel, [The Hitchhiker's Guide to the Galaxy] says, is about the most
    massively useful thing an interstellar hitchhiker can have. Partly it
    has great practical value. You can wrap it around you for warmth as you
    bound across the cold moons of Jaglan Beta; you can lie on it on the brilliant
    marble-sanded beaches of Santraginus V, inhaling the heady sea vapors; you
    can sleep under it beneath the stars which shine so redly on the desert
    world of Kakrafoon; use it to sail a miniraft down the slow heavy
    River Moth; wet it for use in hand-to-hand-combat; wrap it round your
    head to ward off noxious fumes or avoid the gaze of the
    Ravenous Bugblatter Beast of Traal (such a mind-boggingly stupid
    animal, it assumes that if you can't see it, it can't
    see you); you can wave your towel in emergencies as a distress
    signal, and of course dry yourself off with it if it still seems
    to be clean enough.
    """


@pytest.fixture
def long_text(medium_text) -> str:
    return medium_text * 10


@pytest.mark.parametrize(
    "input, expected",
    [
        ("Hello, world!", ["world"]),
        ("This is a test", ["test"]),
        ("", []),
        ("\tThe\n\n\n", []),
    ],
)
def test_no_stop_words_after_tokenization(input, expected):
    assert tokenize(input) == expected


def test_bench_tokenize_small_text(short_text, benchmark):
    result = benchmark(tokenize, short_text)
    assert result


def test_bench_tokenize_medium_text(medium_text, benchmark):
    result = benchmark(tokenize, medium_text)
    assert result


def test_bench_tokenize_long_text(long_text, benchmark):
    result = benchmark(tokenize, long_text)
    assert result



import asyncio
import os
import subprocess
from collections import Counter
from pathlib import Path
import aiohttp
import rich_click as click
from rich.console import Console
from rich.status import Status
from rich.table import Table
from sqlalchemy import create_engine, exc, exists, text
from sqlalchemy.orm import Session
from housaku.db import init_db
from housaku.feeds import fetch_feed, fetch_post
from housaku.files import list_files, extract_tokens
from housaku.models import Doc, Posting, Word
from housaku.repositories import DocRepository, PostingRepository, WordRepository
from housaku.search import search
from housaku.settings import Settings, config_file_path
from housaku.utils import tokenize

console = Console()


@click.group()
@click.pass_context
def cli(ctx) -> None:
    """
    A minimalist personal search engine.
    """

    ctx.ensure_object(dict)

    # Load settings
    settings = Settings()
    ctx.obj["settings"] = settings

    # Setup SQLite database
    engine = create_engine(settings.sqlite_url)
    init_db(engine)

    ctx.obj["engine"] = engine


async def _index_files(
    session: Session,
    status: Status,
    include: list[Path],
    exclude: list[str],
) -> None:
    """
    Index files from the specified list of directories.
    """

    doc_repo = DocRepository(session)
    posting_repo = PostingRepository(session)
    word_repo = WordRepository(session)

    sempahore = asyncio.Semaphore(10)

    async def process_file(file: Path):
        async with sempahore:
            uri = f"{file}"

            try:
                doc_in_db = doc_repo.get_by_attributes(uri=uri)
                if doc_in_db:
                    console.print(
                        f"[bold yellow]SKIP:[/] file already indexed: '{uri}'"
                    )
                    return

                tokens, metadata = await extract_tokens(file)
                doc_in_db = doc_repo.add(Doc(uri=uri, properties=metadata))
                count = Counter(tokens)

                for token, count in count.items():
                    word_in_db = word_repo.get_by_attributes(word=token)
                    if not word_in_db:
                        word_in_db = word_repo.add(Word(word=token))

                    posting_in_db = posting_repo.get_by_attributes(
                        word_id=word_in_db.id,
                        doc_id=doc_in_db.id,
                    )

                    if not posting_in_db:
                        posting_in_db = posting_repo.add(
                            Posting(
                                word=word_in_db,
                                doc=doc_in_db,
                                count=count,
                                tf=count / len(tokens),
                            )
                        )

                console.print(f"[bold green]OK:[/] indexed '{file}'.")
                session.commit()
            except Exception as e:
                console.print(
                    f"[bold red]ERROR:[/] something went wrong while reading '{file}': {e}"
                )

    for dir in include:
        status.update(
            f"[bold green]Indexing documents from '{dir.name}'... Please wait, this may take a moment."
        )
        try:
            files = list_files(dir, exclude)
        except Exception as e:
            console.print(f"[bold red]ERROR:[/] {e}")
            continue

        tasks = [process_file(file) for file in files]
        await asyncio.gather(*tasks)


async def _index_feeds(session: Session, status: Status, feeds: list[str]) -> None:
    """
    Index content from the specified RSS feeds' posts.
    """

    doc_repo = DocRepository(session)
    posting_repo = PostingRepository(session)
    word_repo = WordRepository(session)

    async def process_feed(client: aiohttp.ClientSession, feed_url: str):
        try:
            entries = await fetch_feed(client, feed_url)
            for entry in entries:
                entry_link = entry.link
                uri = f"{entry_link}"

                doc_in_db = doc_repo.get_by_attributes(uri=uri)
                if doc_in_db:
                    console.print(
                        f"[bold yellow]SKIP:[/] post already indexed: '{uri}'"
                    )
                    return

                content = await fetch_post(client, entry_link)
                result = tokenize(content)
                metadata = {
                    "title": entry.get("title", ""),
                    "link": entry_link,
                    "published": entry.get("published", ""),
                    "author": entry.get("author", ""),
                    "summary": entry.get("summary", ""),
                    "categories": [
                        category.term for category in entry.get("categories", [])
                    ],
                }

                doc_in_db = doc_repo.add(Doc(uri=uri, properties=metadata))
                tokens = Counter(result)

                for token, count in tokens.items():
                    word_in_db = word_repo.get_by_attributes(word=token)
                    if not word_in_db:
                        word_in_db = word_repo.add(Word(word=token))

                    posting_in_db = posting_repo.get_by_attributes(
                        word_id=word_in_db.id,
                        doc_id=doc_in_db.id,
                    )

                    if not posting_in_db:
                        posting_in_db = posting_repo.add(
                            Posting(
                                word=word_in_db,
                                doc=doc_in_db,
                                count=count,
                                tf=count / len(result),
                            )
                        )

                console.print(f"[bold green]OK:[/] indexed '{uri}'.")
                session.commit()
        except Exception as e:
            console.print(f"[bold red]ERROR:[/] {e}")

    status.update(
        "[bold green]Indexing feeds and posts... Please wait, this may take a moment."
    )
    async with aiohttp.ClientSession() as client:
        tasks = [process_feed(client, feed) for feed in feeds]
        await asyncio.gather(*tasks)


async def _index(
    session: Session, include: list[Path], exclude: list[str], urls: list[str]
):
    """
    Start indexing of files and feeds based on the settings.
    """

    with console.status(
        "[bold green]Start indexing... Please, wait a moment.",
        spinner="arrow",
    ) as status:
        await _index_files(session, status, include, exclude)
        await _index_feeds(session, status, urls)


@cli.command(
    name="index",
    short_help="Start indexing documents and posts.",
)
@click.option(
    "-i",
    "--include",
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
    ),
    multiple=True,
    help="Directory to include for indexing.",
)
@click.option(
    "-e",
    "--exclude",
    multiple=True,
    help="Directory or pattern to exclude from indexing.",
)
@click.pass_context
def index(ctx, include, exclude) -> None:
    """
    Index documents and posts from the specified sources in the
    config.toml file.
    """

    engine = ctx.obj["engine"]
    settings = ctx.obj["settings"]

    merged_include = list(
        set(Path(d) for d in settings.files.include) | {Path(d) for d in include}
    )
    merged_exclude = list(set(settings.files.exclude) | set(exclude))
    urls = settings.feeds.urls

    with Session(engine) as session:
        asyncio.run(
            _index(
                session,
                merged_include,
                merged_exclude,
                urls,
            )
        )


@cli.command(
    name="search",
    short_help="Search for documents and posts.",
)
@click.option(
    "-q",
    "--query",
    prompt="Search terms",
    help="Search terms to find relevant documents.",
)
@click.option(
    "-l",
    "--limit",
    type=int,
    default=None,
    help="Limit the number of documents returned.",
)
@click.pass_context
def search_documents(ctx, query: str, limit: int) -> None:
    """
    Search for documents and posts based on the provided query.
    """

    engine = ctx.obj["engine"]

    tokens = tokenize(query)
    with Session(engine) as session:
        results = search(session, tokens, limit)
        if not results:
            console.print("[yellow]No results found.[/]")
            return

        table = Table(
            show_header=True,
            header_style="bold",
            show_lines=True,
        )
        table.add_column("Type", width=5, justify="center")
        table.add_column("Name")
        table.add_column("URI")

        for result in results:
            if result.properties.get("name", None):
                file_name = result.properties.get("name", result.uri)
                table.add_row(
                    ":scroll:",
                    file_name,
                    f"[link=file://{result.uri}]{result.uri}[/]",
                )
            else:
                title = result.properties.get("title", result.uri)
                table.add_row(
                    ":globe_with_meridians:",
                    title,
                    f"[link={result.uri}]{result.uri}[/]",
                )

        console.print(table)


@cli.command(name="config", help="Open the configuration file.")
def config() -> None:
    user_editor = os.environ.get("EDITOR", None)
    try:
        if user_editor:
            subprocess.run([user_editor, str(config_file_path)], check=True)
        else:
            subprocess.run(["open", str(config_file_path)], check=True)
    except Exception as e:
        console.print(f"[bold red]ERROR:[/] Failed to open the configuration file: {e}")


@cli.command(
    name="vacuum",
    help="Reclaim unused spaced in the database.",
)
@click.pass_context
def vacuum(ctx) -> None:
    """
    Reclaims unused space in the database.
    """

    engine = ctx.obj["engine"]
    with Session(engine) as session:
        try:
            session.execute(text("VACUUM"))
            session.commit()
            console.print("[bold green]OK:[/] unused space has been reclaimed!")
        except exc.SQLAlchemyError as e:
            console.print(
                f"[bold red]ERROR:[/] something went wrong while reclaiming space: {e}"
            )



import os
from datetime import datetime
from pathlib import Path
from typing import Any, Generator
import fnmatch
import mimetypes
import warnings
from aiofile import async_open
from ebooklib import epub
from pypdf import PdfReader
from docx import Document
import ebooklib
import frontmatter
from housaku.utils import tokenize

# Supress ebooklib warnings.
warnings.filterwarnings("ignore", category=FutureWarning)


def list_files(root: Path, exclude: list[str] = []) -> Generator[Path, Any, Any]:
    if not root.is_dir():
        raise Exception(f"path '{root}' is not a directory")

    exclude_set = set(exclude)
    pending_dirs = [root]
    while pending_dirs:
        dir = pending_dirs.pop()
        for path in dir.iterdir():
            if any(fnmatch.fnmatch(path.name, pattern) for pattern in exclude_set):
                continue

            if path.is_dir():
                pending_dirs.append(path)

            if path.is_file():
                yield path.resolve()


async def extract_tokens(file: Path) -> tuple[list[str], dict]:
    mime_type, _ = mimetypes.guess_type(file)

    if mime_type == "text/plain":
        return await read_txt(file)
    if mime_type == "text/markdown":
        return await read_md(file)
    if mime_type == "application/pdf":
        return await read_pdf(file)
    if mime_type == "application/epub+zip":
        return await read_epub(file)
    if (
        mime_type
        == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    ):
        return await read_docx(file)
    else:
        raise Exception(f"Unsupported file format {mime_type}")


def get_file_metadata(file: Path) -> dict:
    file_stats = os.stat(file)
    return {
        "name": file.name,
        "size": file_stats.st_size,
        "created_at": datetime.fromtimestamp(file_stats.st_ctime).isoformat(),
        "last_modified": datetime.fromtimestamp(file_stats.st_mtime).isoformat(),
    }


async def read_txt(file: Path) -> tuple[list[str], dict]:
    metadata = get_file_metadata(file)

    async with async_open(file, "r") as af:
        content = await af.read()
        return (tokenize(content), metadata)


async def read_md(file: Path) -> tuple[list[str], dict]:
    metadata = get_file_metadata(file)

    async with async_open(file, "r") as af:
        content = await af.read()
        post = frontmatter.loads(content)
        return (
            tokenize(post.content),
            {**metadata, **{key: str(value) for key, value in post.metadata.items()}},
        )


async def read_pdf(file: Path) -> tuple[list[str], dict]:
    metadata = get_file_metadata(file)

    tokens = []
    reader = PdfReader(file.resolve())
    for page in reader.pages:
        tokens.extend(tokenize(page.extract_text()))

    return (tokens, {**metadata, **reader.metadata})


async def read_epub(file: Path) -> tuple[list[str], dict]:
    metadata = get_file_metadata(file)

    tokens = []
    book = epub.read_epub(file, {"ignore_ncx": True})
    for page in book.get_items_of_type(ebooklib.ITEM_DOCUMENT):
        content = page.get_body_content().decode()
        tokens.extend(tokenize(content))

    book_metadata = {
        field: (
            book.get_metadata("DC", field)[0][0]
            if book.get_metadata("DC", field)
            else None
        )
        for field in ["title", "creator", "identifier"]
    }

    return (tokens, {**metadata, **book_metadata})


async def read_docx(file: Path) -> tuple[list[str], dict]:
    metadata = get_file_metadata(file)

    tokens = []
    document = Document(f"{file}")
    for paragraph in document.paragraphs:
        tokens.extend(tokenize(paragraph.text))

    return (tokens, metadata)



from sqlalchemy import update
from sqlalchemy.orm import Session
from housaku.models import Doc, Posting, Word


class Repository[T: (Doc, Posting, Word)]:
    def __init__(self, session: Session, model) -> None:
        self.session = session
        self.model = model

    def add(self, entity: T) -> T:
        self.session.add(entity)
        self.session.commit()
        self.session.refresh(entity)

        return entity

    def add_in_batch(self, entities: list[T]) -> None:
        self.session.add_all(entities)
        self.session.commit()

    def get(self, id: int) -> T | None:
        return self.session.get(self.model, id)

    def get_by_attributes(self, **kwargs) -> T | None:
        return self.session.query(self.model).filter_by(**kwargs).one_or_none()

    def get_all(self) -> list[T]:
        return self.session.query(self.model).all()

    def get_all_by_attributes(self, **kwargs) -> list[T] | None:
        return self.session.query(self.model).filter_by(**kwargs).all()

    def update(self, id: int, **kwargs) -> None:
        stmt = update(self.model).where(self.model.id == id).values(**kwargs)
        self.session.execute(stmt)
        self.session.commit()

    def delete(self, id: int) -> None:
        entity = self.session.get(self.model, id)
        if entity:
            self.session.delete(entity)
            self.session.commit()
            return entity


class DocRepository(Repository[Doc]):
    def __init__(self, session: Session) -> None:
        super().__init__(session, Doc)


class PostingRepository(Repository[Posting]):
    def __init__(self, session: Session) -> None:
        super().__init__(session, Posting)


class WordRepository(Repository[Word]):
    def __init__(self, session: Session) -> None:
        super().__init__(session, Word)



import string
from selectolax.parser import HTMLParser
from housaku.stop_words import STOP_WORDS

translation_table = str.maketrans(string.punctuation, " " * len(string.punctuation))


def tokenize(text: str) -> list[str]:
    tokens = text.lower().translate(translation_table).split()
    tokens = [
        token for token in tokens if token not in STOP_WORDS and not token[0].isdigit()
    ]
    return tokens


def clean_html(html: str) -> str:
    tree = HTMLParser(html)
    for tag in tree.css("script, style, video, img, canvas"):
        tag.decompose()

    text = "".join(node.text(deep=True) for node in tree.css("main"))
    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split(" "))
    cleaned_text = " ".join(chunk for chunk in chunks if chunk)

    return cleaned_text



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from housaku.db import init_db
from housaku.models import Base, Doc, Posting, Word
from housaku.repositories import DocRepository, PostingRepository, WordRepository
from housaku.search import search


@pytest.fixture(scope="function")
def session():
    engine = create_engine("sqlite:///:memory:")
    init_db(engine)

    Base.metadata.create_all(engine)

    with Session(engine) as session:
        yield session


@pytest.fixture
def doc_repo(session):
    return DocRepository(session)


@pytest.fixture
def posting_repo(session):
    return PostingRepository(session)


@pytest.fixture
def word_repo(session):
    return WordRepository(session)


def test_search_no_docs(session):
    result = search(session, ["tdd"])
    assert result == []


def test_search_no_matching_words(session, doc_repo):
    doc_repo.add(Doc(uri="test1"))
    doc_repo.add(Doc(uri="test2"))
    tokens = ["nonexistent", "words"]

    results = search(session, tokens)
    assert results == []


def test_search_empty_token_list(session, doc_repo):
    doc_repo.add(Doc(uri="test1"))
    doc_repo.add(Doc(uri="test2"))

    results = search(session, [])
    assert results == []


def test_search_two_docs(session, doc_repo, posting_repo, word_repo):
    doc1_in_db = doc_repo.add(Doc(uri="test1"))
    doc_repo.add(Doc(uri="test2"))
    tokens = ["tdd", "does", "not", "work"]

    for token in tokens:
        word_in_db = word_repo.add(Word(word=token))
        posting_repo.add(
            Posting(
                word=word_in_db,
                doc=doc1_in_db,
                count=1,
                tf=1 / len(tokens),
            )
        )

    results = search(session, ["tdd"])
    assert len(results) == 1
    assert results[0].id == doc1_in_db.id


def test_search_different_tf_values(session, doc_repo, posting_repo, word_repo):
    doc1_in_db = doc_repo.add(Doc(uri="test1"))
    doc2_in_db = doc_repo.add(Doc(uri="test2"))
    token = "tdd"

    word_in_db = word_repo.add(Word(word=token))
    posting_repo.add(
        Posting(
            word=word_in_db,
            doc=doc1_in_db,
            count=3,
            tf=3 / 5,
        )
    )
    posting_repo.add(
        Posting(
            word=word_in_db,
            doc=doc2_in_db,
            count=1,
            tf=1 / 5,
        )
    )

    results = search(session, [token])
    assert len(results) == 2
    assert results[0].id == doc1_in_db.id
    assert results[1].id == doc2_in_db.id



from sqlalchemy import JSON, Float, ForeignKey, Text, Integer
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship


class Base(DeclarativeBase):
    pass


class Doc(Base):
    __tablename__ = "documents"

    id: Mapped[int] = mapped_column(primary_key=True)
    uri: Mapped[str] = mapped_column(nullable=False, unique=True)
    content: Mapped[str] = mapped_column(Text, nullable=True)
    properties: Mapped[dict] = mapped_column(JSON, nullable=True)

    postings: Mapped[list["Posting"]] = relationship(back_populates="doc")

    def __repr__(self) -> str:
        return f"Doc(id={self.id!r}, uri={self.uri!r})"


class Posting(Base):
    __tablename__ = "postings"

    id: Mapped[int] = mapped_column(primary_key=True)
    count: Mapped[int] = mapped_column(Integer, default=1, nullable=False)
    tf: Mapped[float] = mapped_column(Float, default=0, nullable=False)

    doc_id: Mapped[int] = mapped_column(ForeignKey("documents.id"), index=True)
    doc: Mapped["Doc"] = relationship(back_populates="postings")

    word_id: Mapped[int] = mapped_column(ForeignKey("words.id"), index=True)
    word: Mapped["Word"] = relationship(back_populates="postings")

    def __repr__(self) -> str:
        return f"Posting(id={self.id!r}, doc_id={self.doc_id!r}, word_id={self.word_id!r}, count={self.count!r}, tf={self.tf!r})"


class Word(Base):
    __tablename__ = "words"

    id: Mapped[int] = mapped_column(primary_key=True)
    word: Mapped[str] = mapped_column(
        nullable=False,
        unique=True,
        index=True,
    )

    postings: Mapped[list["Posting"]] = relationship(back_populates="word")

    def __repr__(self) -> str:
        return (
            f"Word(id={self.id!r}, word={self.word!r}, postings={len(self.postings)!r})"
        )



STOP_WORDS = {
    "a",
    "a's",
    "abbia",
    "abbiamo",
    "abbiano",
    "abbiate",
    "able",
    "about",
    "above",
    "according",
    "accordingly",
    "across",
    "actually",
    "ad",
    "after",
    "afterwards",
    "again",
    "against",
    "agl",
    "agli",
    "ai",
    "aie",
    "aient",
    "aies",
    "ain't",
    "ait",
    "al",
    "algo",
    "algunas",
    "algunos",
    "all",
    "alla",
    "alle",
    "allo",
    "allow",
    "allows",
    "almost",
    "alone",
    "along",
    "already",
    "also",
    "although",
    "always",
    "am",
    "among",
    "amongst",
    "an",
    "anche",
    "and",
    "another",
    "ante",
    "antes",
    "any",
    "anybody",
    "anyhow",
    "anyone",
    "anything",
    "anyway",
    "anyways",
    "anywhere",
    "ao",
    "aos",
    "apart",
    "appear",
    "appreciate",
    "appropriate",
    "aquela",
    "aquelas",
    "aquele",
    "aqueles",
    "aquilo",
    "are",
    "aren't",
    "around",
    "as",
    "aside",
    "ask",
    "asking",
    "associated",
    "at",
    "até",
    "au",
    "aura",
    "aurai",
    "auraient",
    "aurais",
    "aurait",
    "auras",
    "aurez",
    "auriez",
    "aurions",
    "aurons",
    "auront",
    "aux",
    "avaient",
    "available",
    "avais",
    "avait",
    "avec",
    "avemmo",
    "avendo",
    "avesse",
    "avessero",
    "avessi",
    "avessimo",
    "aveste",
    "avesti",
    "avete",
    "aveva",
    "avevamo",
    "avevano",
    "avevate",
    "avevi",
    "avevo",
    "avez",
    "aviez",
    "avions",
    "avons",
    "avrai",
    "avranno",
    "avrebbe",
    "avrebbero",
    "avrei",
    "avremmo",
    "avremo",
    "avreste",
    "avresti",
    "avrete",
    "avrà",
    "avrò",
    "avuta",
    "avute",
    "avuti",
    "avuto",
    "away",
    "awfully",
    "ayant",
    "ayez",
    "ayons",
    "b",
    "be",
    "became",
    "because",
    "become",
    "becomes",
    "becoming",
    "been",
    "before",
    "beforehand",
    "behind",
    "being",
    "believe",
    "below",
    "beside",
    "besides",
    "best",
    "better",
    "between",
    "beyond",
    "both",
    "brief",
    "but",
    "by",
    "c",
    "c'mon",
    "c's",
    "came",
    "can",
    "can't",
    "cannot",
    "cant",
    "cause",
    "causes",
    "ce",
    "ceci",
    "cela",
    "celà",
    "certain",
    "certainly",
    "ces",
    "cet",
    "cette",
    "changes",
    "che",
    "chi",
    "ci",
    "clearly",
    "co",
    "coi",
    "col",
    "com",
    "come",
    "comes",
    "como",
    "con",
    "concerning",
    "consequently",
    "consider",
    "considering",
    "contain",
    "containing",
    "contains",
    "contra",
    "contro",
    "corresponding",
    "could",
    "couldn't",
    "course",
    "cual",
    "cuando",
    "cui",
    "currently",
    "d",
    "da",
    "dagl",
    "dagli",
    "dai",
    "dal",
    "dall",
    "dalla",
    "dalle",
    "dallo",
    "dans",
    "das",
    "de",
    "definitely",
    "degl",
    "degli",
    "dei",
    "del",
    "dela",
    "delas",
    "dele",
    "deles",
    "dell",
    "della",
    "delle",
    "dello",
    "depois",
    "des",
    "described",
    "desde",
    "despite",
    "di",
    "did",
    "didn't",
    "different",
    "do",
    "does",
    "doesn't",
    "doing",
    "don't",
    "donde",
    "done",
    "dos",
    "dov",
    "dove",
    "down",
    "downwards",
    "du",
    "durante",
    "during",
    "e",
    "each",
    "ebbe",
    "ebbero",
    "ebbi",
    "ed",
    "edu",
    "eg",
    "eight",
    "either",
    "el",
    "ela",
    "elas",
    "ele",
    "eles",
    "ella",
    "ellas",
    "elle",
    "ellos",
    "else",
    "elsewhere",
    "em",
    "en",
    "enough",
    "entirely",
    "entre",
    "era",
    "erais",
    "eram",
    "eran",
    "erano",
    "eras",
    "eravamo",
    "eravate",
    "eres",
    "eri",
    "ero",
    "es",
    "esa",
    "esas",
    "ese",
    "eso",
    "esos",
    "especially",
    "essa",
    "essas",
    "esse",
    "essendo",
    "esses",
    "est",
    "esta",
    "estaba",
    "estabais",
    "estaban",
    "estabas",
    "estada",
    "estadas",
    "estado",
    "estados",
    "estamos",
    "estando",
    "estar",
    "estaremos",
    "estará",
    "estarán",
    "estarás",
    "estaré",
    "estaréis",
    "estaría",
    "estaríais",
    "estaríamos",
    "estarían",
    "estarías",
    "estas",
    "estava",
    "estavam",
    "este",
    "esteja",
    "estejam",
    "estejamos",
    "estemos",
    "estes",
    "esteve",
    "estive",
    "estivemos",
    "estiver",
    "estivera",
    "estiveram",
    "estiverem",
    "estivermos",
    "estivesse",
    "estivessem",
    "estivéramos",
    "estivéssemos",
    "esto",
    "estos",
    "estou",
    "estoy",
    "estuve",
    "estuviera",
    "estuvierais",
    "estuvieran",
    "estuvieras",
    "estuvieron",
    "estuviese",
    "estuvieseis",
    "estuviesen",
    "estuvieses",
    "estuvimos",
    "estuviste",
    "estuvisteis",
    "estuviéramos",
    "estuviésemos",
    "estuvo",
    "está",
    "estábamos",
    "estáis",
    "están",
    "estás",
    "estávamos",
    "estão",
    "esté",
    "estéis",
    "estén",
    "estés",
    "et",
    "etc",
    "eu",
    "eue",
    "eues",
    "eurent",
    "eus",
    "eusse",
    "eussent",
    "eusses",
    "eussiez",
    "eussions",
    "eut",
    "eux",
    "even",
    "ever",
    "every",
    "everybody",
    "everyone",
    "everything",
    "everywhere",
    "ex",
    "exactly",
    "example",
    "except",
    "eûmes",
    "eût",
    "eûtes",
    "f",
    "faccia",
    "facciamo",
    "facciano",
    "facciate",
    "faccio",
    "facemmo",
    "facendo",
    "facesse",
    "facessero",
    "facessi",
    "facessimo",
    "faceste",
    "facesti",
    "faceva",
    "facevamo",
    "facevano",
    "facevate",
    "facevi",
    "facevo",
    "fai",
    "fanno",
    "far",
    "farai",
    "faranno",
    "farebbe",
    "farebbero",
    "farei",
    "faremmo",
    "faremo",
    "fareste",
    "faresti",
    "farete",
    "farà",
    "farò",
    "fece",
    "fecero",
    "feci",
    "few",
    "fifth",
    "first",
    "five",
    "foi",
    "followed",
    "following",
    "follows",
    "fomos",
    "for",
    "fora",
    "foram",
    "forem",
    "former",
    "formerly",
    "formos",
    "forth",
    "fosse",
    "fossem",
    "fossero",
    "fossi",
    "fossimo",
    "foste",
    "fosti",
    "four",
    "from",
    "fu",
    "fue",
    "fuera",
    "fuerais",
    "fueran",
    "fueras",
    "fueron",
    "fuese",
    "fueseis",
    "fuesen",
    "fueses",
    "fui",
    "fuimos",
    "fuiste",
    "fuisteis",
    "fummo",
    "furent",
    "furono",
    "further",
    "furthermore",
    "fus",
    "fusse",
    "fussent",
    "fusses",
    "fussiez",
    "fussions",
    "fut",
    "fuéramos",
    "fuésemos",
    "fôramos",
    "fôssemos",
    "fûmes",
    "fût",
    "fûtes",
    "g",
    "get",
    "gets",
    "getting",
    "given",
    "gives",
    "gli",
    "go",
    "goes",
    "going",
    "gone",
    "got",
    "gotten",
    "greetings",
    "h",
    "ha",
    "haber",
    "habida",
    "habidas",
    "habido",
    "habidos",
    "habiendo",
    "habremos",
    "habrá",
    "habrán",
    "habrás",
    "habré",
    "habréis",
    "habría",
    "habríais",
    "habríamos",
    "habrían",
    "habrías",
    "habéis",
    "había",
    "habíais",
    "habíamos",
    "habían",
    "habías",
    "had",
    "hadn't",
    "hai",
    "haja",
    "hajam",
    "hajamos",
    "han",
    "hanno",
    "happens",
    "hardly",
    "has",
    "hasn't",
    "hasta",
    "have",
    "havemos",
    "haven't",
    "havia",
    "having",
    "hay",
    "haya",
    "hayamos",
    "hayan",
    "hayas",
    "hayáis",
    "he",
    "he's",
    "hei",
    "hello",
    "help",
    "hemos",
    "hence",
    "her",
    "here",
    "here's",
    "hereafter",
    "hereby",
    "herein",
    "hereupon",
    "hers",
    "herself",
    "hi",
    "him",
    "himself",
    "his",
    "hither",
    "ho",
    "hopefully",
    "houve",
    "houvemos",
    "houver",
    "houvera",
    "houveram",
    "houverei",
    "houverem",
    "houveremos",
    "houveria",
    "houveriam",
    "houvermos",
    "houverá",
    "houverão",
    "houveríamos",
    "houvesse",
    "houvessem",
    "houvéramos",
    "houvéssemos",
    "how",
    "howbeit",
    "however",
    "hube",
    "hubiera",
    "hubierais",
    "hubieran",
    "hubieras",
    "hubieron",
    "hubiese",
    "hubieseis",
    "hubiesen",
    "hubieses",
    "hubimos",
    "hubiste",
    "hubisteis",
    "hubiéramos",
    "hubiésemos",
    "hubo",
    "há",
    "hão",
    "i",
    "i'd",
    "i'll",
    "i'm",
    "i've",
    "ici",
    "ie",
    "if",
    "ignored",
    "il",
    "ils",
    "immediate",
    "in",
    "inasmuch",
    "inc",
    "indeed",
    "indicate",
    "indicated",
    "indicates",
    "inner",
    "insofar",
    "instead",
    "into",
    "inward",
    "io",
    "is",
    "isn't",
    "isso",
    "isto",
    "it",
    "it'd",
    "it'll",
    "it's",
    "its",
    "itself",
    "j",
    "je",
    "just",
    "já",
    "k",
    "keep",
    "keeps",
    "kept",
    "know",
    "known",
    "knows",
    "l",
    "la",
    "las",
    "last",
    "lately",
    "later",
    "latter",
    "latterly",
    "le",
    "least",
    "lei",
    "les",
    "less",
    "lest",
    "let",
    "let's",
    "leur",
    "leurs",
    "lhe",
    "lhes",
    "li",
    "like",
    "liked",
    "likely",
    "little",
    "lo",
    "look",
    "looking",
    "looks",
    "loro",
    "los",
    "ltd",
    "lui",
    "m",
    "ma",
    "mainly",
    "mais",
    "many",
    "mas",
    "may",
    "maybe",
    "me",
    "mean",
    "meanwhile",
    "merely",
    "mes",
    "mesmo",
    "meu",
    "meus",
    "mi",
    "mia",
    "mie",
    "miei",
    "might",
    "minha",
    "minhas",
    "mio",
    "mis",
    "moi",
    "mon",
    "more",
    "moreover",
    "most",
    "mostly",
    "much",
    "mucho",
    "muchos",
    "muito",
    "must",
    "muy",
    "my",
    "myself",
    "más",
    "même",
    "mí",
    "mía",
    "mías",
    "mío",
    "míos",
    "n",
    "na",
    "nada",
    "name",
    "namely",
    "nas",
    "nd",
    "ne",
    "near",
    "nearly",
    "necessary",
    "need",
    "needs",
    "negl",
    "negli",
    "nei",
    "neither",
    "nel",
    "nell",
    "nella",
    "nelle",
    "nello",
    "nem",
    "never",
    "nevertheless",
    "new",
    "next",
    "ni",
    "nine",
    "no",
    "nobody",
    "noi",
    "non",
    "none",
    "noone",
    "nor",
    "normally",
    "nos",
    "nosotras",
    "nosotros",
    "nossa",
    "nossas",
    "nosso",
    "nossos",
    "nostra",
    "nostre",
    "nostri",
    "nostro",
    "not",
    "nothing",
    "notre",
    "nous",
    "novel",
    "now",
    "nowhere",
    "nuestra",
    "nuestras",
    "nuestro",
    "nuestros",
    "num",
    "numa",
    "não",
    "nós",
    "o",
    "obviously",
    "of",
    "off",
    "often",
    "oh",
    "ok",
    "okay",
    "old",
    "on",
    "once",
    "one",
    "ones",
    "only",
    "ont",
    "onto",
    "or",
    "os",
    "other",
    "others",
    "otherwise",
    "otra",
    "otras",
    "otro",
    "otros",
    "ou",
    "ought",
    "our",
    "ours",
    "ourselves",
    "out",
    "outside",
    "over",
    "overall",
    "own",
    "p",
    "par",
    "para",
    "particular",
    "particularly",
    "pas",
    "pela",
    "pelas",
    "pelo",
    "pelos",
    "per",
    "perché",
    "perhaps",
    "pero",
    "più",
    "placed",
    "please",
    "plus",
    "poco",
    "por",
    "porque",
    "possible",
    "pour",
    "presumably",
    "probably",
    "provides",
    "q",
    "qu",
    "qual",
    "quale",
    "quando",
    "quanta",
    "quante",
    "quanti",
    "quanto",
    "que",
    "quel",
    "quella",
    "quelle",
    "quelles",
    "quelli",
    "quello",
    "quels",
    "quem",
    "questa",
    "queste",
    "questi",
    "questo",
    "qui",
    "quien",
    "quienes",
    "quite",
    "qué",
    "qv",
    "r",
    "rather",
    "rd",
    "re",
    "really",
    "reasonably",
    "regarding",
    "regardless",
    "regards",
    "relatively",
    "respectively",
    "right",
    "s",
    "sa",
    "said",
    "same",
    "sans",
    "sarai",
    "saranno",
    "sarebbe",
    "sarebbero",
    "sarei",
    "saremmo",
    "saremo",
    "sareste",
    "saresti",
    "sarete",
    "sarà",
    "sarò",
    "saw",
    "say",
    "saying",
    "says",
    "se",
    "sea",
    "seamos",
    "sean",
    "seas",
    "second",
    "secondly",
    "see",
    "seeing",
    "seem",
    "seemed",
    "seeming",
    "seems",
    "seen",
    "sei",
    "seja",
    "sejam",
    "sejamos",
    "self",
    "selves",
    "sem",
    "sensible",
    "sent",
    "ser",
    "sera",
    "serai",
    "seraient",
    "serais",
    "serait",
    "seras",
    "serei",
    "seremos",
    "serez",
    "seria",
    "seriam",
    "seriez",
    "serions",
    "serious",
    "seriously",
    "serons",
    "seront",
    "será",
    "serán",
    "serás",
    "serão",
    "seré",
    "seréis",
    "sería",
    "seríais",
    "seríamos",
    "serían",
    "serías",
    "ses",
    "seu",
    "seus",
    "seven",
    "several",
    "seáis",
    "shall",
    "she",
    "should",
    "shouldn't",
    "si",
    "sia",
    "siamo",
    "siano",
    "siate",
    "sido",
    "siendo",
    "siete",
    "sin",
    "since",
    "six",
    "so",
    "sobre",
    "soi",
    "soient",
    "sois",
    "soit",
    "some",
    "somebody",
    "somehow",
    "someone",
    "something",
    "sometime",
    "sometimes",
    "somewhat",
    "somewhere",
    "sommes",
    "somos",
    "son",
    "sono",
    "sont",
    "soon",
    "sorry",
    "sou",
    "soy",
    "soyez",
    "soyons",
    "specified",
    "specify",
    "specifying",
    "sta",
    "stai",
    "stando",
    "stanno",
    "starai",
    "staranno",
    "starebbe",
    "starebbero",
    "starei",
    "staremmo",
    "staremo",
    "stareste",
    "staresti",
    "starete",
    "starà",
    "starò",
    "stava",
    "stavamo",
    "stavano",
    "stavate",
    "stavi",
    "stavo",
    "stemmo",
    "stesse",
    "stessero",
    "stessi",
    "stessimo",
    "steste",
    "stesti",
    "stette",
    "stettero",
    "stetti",
    "stia",
    "stiamo",
    "stiano",
    "stiate",
    "still",
    "sto",
    "su",
    "sua",
    "suas",
    "sub",
    "such",
    "sue",
    "sugl",
    "sugli",
    "sui",
    "suis",
    "sul",
    "sull",
    "sulla",
    "sulle",
    "sullo",
    "suo",
    "suoi",
    "sup",
    "sur",
    "sure",
    "sus",
    "suya",
    "suyas",
    "suyo",
    "suyos",
    "são",
    "sí",
    "só",
    "t",
    "t's",
    "ta",
    "take",
    "taken",
    "también",
    "também",
    "tanto",
    "te",
    "tell",
    "tem",
    "temos",
    "tendremos",
    "tendrá",
    "tendrán",
    "tendrás",
    "tendré",
    "tendréis",
    "tendría",
    "tendríais",
    "tendríamos",
    "tendrían",
    "tendrías",
    "tends",
    "tened",
    "tenemos",
    "tenga",
    "tengamos",
    "tengan",
    "tengas",
    "tengo",
    "tengáis",
    "tenha",
    "tenham",
    "tenhamos",
    "tenho",
    "tenida",
    "tenidas",
    "tenido",
    "tenidos",
    "teniendo",
    "tenéis",
    "tenía",
    "teníais",
    "teníamos",
    "tenían",
    "tenías",
    "ter",
    "terei",
    "teremos",
    "teria",
    "teriam",
    "terá",
    "terão",
    "teríamos",
    "tes",
    "teu",
    "teus",
    "teve",
    "th",
    "than",
    "thank",
    "thanks",
    "thanx",
    "that",
    "that's",
    "thats",
    "the",
    "their",
    "theirs",
    "them",
    "themselves",
    "then",
    "thence",
    "there",
    "there's",
    "thereafter",
    "thereby",
    "therefore",
    "therein",
    "theres",
    "thereupon",
    "these",
    "they",
    "they'd",
    "they'll",
    "they're",
    "they've",
    "think",
    "third",
    "this",
    "thorough",
    "thoroughly",
    "those",
    "though",
    "three",
    "through",
    "throughout",
    "thru",
    "thus",
    "ti",
    "tiene",
    "tienen",
    "tienes",
    "tinha",
    "tinham",
    "tive",
    "tivemos",
    "tiver",
    "tivera",
    "tiveram",
    "tiverem",
    "tivermos",
    "tivesse",
    "tivessem",
    "tivéramos",
    "tivéssemos",
    "to",
    "todo",
    "todos",
    "together",
    "toi",
    "ton",
    "too",
    "took",
    "toward",
    "towards",
    "tra",
    "tried",
    "tries",
    "truly",
    "try",
    "trying",
    "tu",
    "tua",
    "tuas",
    "tue",
    "tuo",
    "tuoi",
    "tus",
    "tutti",
    "tutto",
    "tuve",
    "tuviera",
    "tuvierais",
    "tuvieran",
    "tuvieras",
    "tuvieron",
    "tuviese",
    "tuvieseis",
    "tuviesen",
    "tuvieses",
    "tuvimos",
    "tuviste",
    "tuvisteis",
    "tuviéramos",
    "tuviésemos",
    "tuvo",
    "tuya",
    "tuyas",
    "tuyo",
    "tuyos",
    "twice",
    "two",
    "tém",
    "têm",
    "tínhamos",
    "tú",
    "u",
    "um",
    "uma",
    "un",
    "una",
    "under",
    "une",
    "unfortunately",
    "unless",
    "unlikely",
    "uno",
    "unos",
    "until",
    "unto",
    "up",
    "upon",
    "us",
    "use",
    "used",
    "useful",
    "uses",
    "using",
    "usually",
    "uucp",
    "v",
    "value",
    "various",
    "very",
    "vi",
    "via",
    "viz",
    "você",
    "vocês",
    "voi",
    "vos",
    "vosotras",
    "vosotros",
    "vostra",
    "vostre",
    "vostri",
    "vostro",
    "votre",
    "vous",
    "vs",
    "vuestra",
    "vuestras",
    "vuestro",
    "vuestros",
    "w",
    "want",
    "wants",
    "was",
    "wasn't",
    "way",
    "we",
    "we'd",
    "we'll",
    "we're",
    "we've",
    "welcome",
    "well",
    "went",
    "were",
    "weren't",
    "what",
    "what's",
    "whatever",
    "when",
    "whence",
    "whenever",
    "where",
    "where's",
    "whereafter",
    "whereas",
    "whereby",
    "wherein",
    "whereupon",
    "wherever",
    "whether",
    "which",
    "while",
    "whither",
    "who",
    "who's",
    "whoever",
    "whole",
    "whom",
    "whose",
    "why",
    "will",
    "willing",
    "wish",
    "with",
    "within",
    "without",
    "won't",
    "wonder",
    "would",
    "wouldn't",
    "x",
    "y",
    "ya",
    "yes",
    "yet",
    "yo",
    "you",
    "you'd",
    "you'll",
    "you're",
    "you've",
    "your",
    "yours",
    "yourself",
    "yourselves",
    "z",
    "zero",
    "à",
    "às",
    "è",
    "é",
    "él",
    "éramos",
    "étaient",
    "étais",
    "était",
    "étant",
    "étiez",
    "étions",
    "été",
    "étée",
    "étées",
    "étés",
    "êtes",
}
